# Feature implementing core functionality
#
# Copyright (c) 2018 Ensoft Ltd

import logging, os
from .cfg_base import ConfiguredFeature
from .dyn_base import DynamicFeature
from .tgt_base import TargetFeature

log = logging.getLogger(__name__)


class CoreFeature(ConfiguredFeature):
    """
    Feature to manage core services
    """

    #
    # Schema
    #
    name = "core"

    requests = {
        "force_restart": [],
        "ping": [],
        "start_feature": ["feature", "channel", "target", "__req__"],
        "stop_feature": ["feature", "channel", "target"],
    }

    notifications = ["pong"]

    config = {"allow_restart_requests": False, "allowed_dynamic_features": None}

    #
    # Implementation
    #
    def __init__(self, ws_handler, config):
        super().__init__(ws_handler, config)
        self.started_features = {}

    async def do_force_restart(self):
        """
        Forcibly restart the server (assuming started by the "run" script)
        """
        if self.config["allow_restart_requests"]:
            os._exit(42)
        else:
            return self._rpc_failure("Restart disallowed by configuration")

    async def do_ping(self):
        """
        Respond to a keepalive message from the client
        """
        return self._result("pong")

    async def do_start_feature(self, feature_name, channel, target, req):
        """
        Start an optional feature
        """
        allowed = self.config["allowed_dynamic_features"]
        if allowed is not None and feature_name not in allowed:
            return self._rpc_failure(
                "Feature {} is disallowed by configuration".format(feature_name)
            )
        feature_cls = DynamicFeature.find(feature_name)
        new_feature = feature_cls(self.ws_handler, channel, target, req)
        self.ws_handler.add_feature(new_feature, channel, target)
        self.started_features[_fk(feature_name, channel, target)] = new_feature

    async def do_stop_feature(self, feature_name, channel, target):
        """
        Stop an optional feature
        """
        feature_key = _fk(feature_name, channel, target)
        feature = self.started_features[feature_key]
        await self.ws_handler.remove_feature(feature, channel, target)
        try:
            del self.started_features[feature_key]
        except KeyError:
            # probably some race condition
            log.debug(
                "Ignoring missing started feature %s on stop_feature request",
                feature_key,
            )

        if isinstance(feature, TargetFeature):
            # Might as well try a disconnect
            log.debug("About to disconnect stopped feature " + feature_key)
            await feature.disconnect()


def _fk(feature_name, channel, target):
    """Construct a dict key for a feature instance"""
    return "{}::{}::{}".format(feature_name, channel, target)



import runpy

from entrance.connection import *
from entrance.exceptions import *
from entrance.feature import *
from entrance.ws_handler import *


def main():
    """
    Run the entrance module (via __main__.py) in the current Python process.
    """
    runpy.run_module("entrance", run_name="__main__")



# Syslog monitoring feature
#
# Copyright (c) 2018 Ensoft Ltd

import re, time

from .._util import events
from .tgt_base import TargetFeature


class SyslogFeature(TargetFeature):
    """
    Feature that monitors for syslogs, with optional debugs and filtering.
    """

    #
    # Schema
    #
    name = "syslog"
    notifications = ["syslog"]

    #
    # Implementation
    #
    async def connect(self, conn_factory):
        """
        Set up our connection to the router
        """
        self.connection = await conn_factory.get_cli_connection(
            "syslog", finalizer=self.finalizer
        )
        self.add_connection(self.connection, from_scratch=True)

    async def finalizer(self):
        """
        Finalize a new connection
        """
        # First handle any filters/debug settings from the start_feature req
        filters = self.original_req.get("filters", [])
        if len(filters) == 0:
            # just drop empty lines
            filters = [r"\S"]
        regexp = re.compile("|".join(["({})".format(f) for f in filters]))
        nfn = {"nfn_type": "syslog", "channel": self.original_req["channel"]}
        if "id" in self.original_req:
            nfn["id"] = self.original_req["id"]

        # Then do the expect stuff to get us ready to go
        await self.connection.send("undebug all all-tty\n", override=True)
        await self.connection.expect_prompt(override=True)
        await self.connection.send("terminal monitor\n", override=True)
        await self.connection.expect_prompt(override=True)
        for debug in self.original_req.get("debugs", []):
            await self.connection.send("{}\n".format(debug), override=True)
            await self.connection.expect_prompt(override=True)

        # We need to return at this point, so the connection transitions
        # from FINALIZING to CONNECTED, so tee up our mini event loop for later
        events.create_checked_task(self._event_loop(regexp, nfn))

    async def _event_loop(self, regexp, nfn):
        """
        Sit and wait for syslogs/debugs to come in
        """
        # poll event loop for disconnect requests occasionally
        await self.connection.settimeout(1)

        # enter event loop
        while not self.connection.terminate:
            data = (await self.connection.recv()).decode()
            for syslog in data.split("\n"):
                if regexp.search(syslog):
                    nfn["result"] = syslog
                    nfn["time"] = time.time()
                    await self._notify(**nfn)



# EnTrance default application start
#
# Copyright (c) 2018 Ensoft Ltd

import argparse, functools, logging, logging.config, sys
import sanic, sanic.worker.loader, yaml

from . import server
from ._util import logger

log = logging.getLogger(__name__)
location = getattr(sys, "_MEIPASS", ".") + "/"  # where pre-canned files live


def parse_args(args):
    """
    Parse command-line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--port", help="override port")
    parser.add_argument("-a", "--addr", help="override bind address")
    parser.add_argument("-d", "--debug", action="store_true", help="debug to console")
    parser.add_argument(
        "-c", "--config", default=location + "config.yml", help="config file"
    )
    parser.add_argument("-l", "--logging", help="logging definition file")
    return parser.parse_args(args)


def main(*args, task=None):
    # Load up preferences
    opts = parse_args(args)

    # Config file is always required. The logging file may fall back
    # to the default yaml file below if unspecified.
    main_config = yaml.safe_load(open(opts.config).read())
    if opts.logging is None:
        logging_yaml = logging_yaml_default
    else:
        logging_yaml = open(opts.logging).read()
    logging_config = yaml.safe_load(logging_yaml)

    # Apply any command-line overrides
    if opts.port is not None:
        main_config["start"]["port"] = opts.port
    if opts.addr is not None:
        main_config["start"]["host"] = opts.addr
    if opts.debug:
        logging_config["handlers"]["console"]["level"] = "DEBUG"

    # Go
    logging.config.dictConfig(logging_config)
    logging.setLogRecordFactory(logger.FormattedLogRecord)

    loader = sanic.worker.loader.AppLoader(
        factory=functools.partial(server.create_app, main_config, location)
    )
    app = loader.load()

    # Optionally invoke a caller-specified task once the event loop is up
    if task is not None:
        app.add_task(task)

    start_cfg = main_config["start"]
    log.info(
        "Starting app with %s",
        ", ".join(["{}={}".format(k, v) for k, v in start_cfg.items()]),
    )

    # Enter event loop
    app.prepare(host=start_cfg["host"], port=int(start_cfg["port"]), dev=True)
    sanic.Sanic.serve(primary=app, app_loader=loader)

    log.info("Closing down gracefully")


# Default logging.yml file
logging_yaml_default = """
# Separate configuration file for logging, since most end-users care less
# about the details of this
version: 1
disable_existing_loggers: false
formatters:
    brief:
        format: "%(levelname)s: [%(name)s] %(message)s"
    normal:
        format: "%(asctime)s %(levelname)s: [%(name)s] %(message)s"
handlers:
    console:
        class: logging.StreamHandler
        level: INFO
        formatter: brief
        stream: ext://sys.stdout
    debug:
        class: logging.handlers.RotatingFileHandler
        level: DEBUG
        formatter: normal
        filename: debug.log
        maxBytes: 1000000
        backupCount: 3
        encoding: utf8
loggers:
    "":
        level: DEBUG
        handlers: [console, debug]
    asyncio:
        level: DEBUG
        handlers: [debug]
    paramiko:
        level: WARNING
        handlers: [console, debug]
    ncclient:
        level: WARNING
        handlers: [console, debug]
    network:
        level: DEBUG
        handlers: [debug]
    sanic:
        level: WARNING
        handlers: [debug]
    websockets:
        level: WARNING
        handlers: [console, debug]
"""

if __name__ == "__main__":
    main(*sys.argv[1:])



# EnTrance server
#
# Copyright (c) 2023 Ensoft Ltd

import logging
import sys

import sanic
import sanic.response

from . import WebsocketHandler


log = logging.getLogger(__name__)


def create_app(config, file_location) -> sanic.Sanic:
    """
    Create a simple server with the specified configuration.

    If an app needs more elaborate setup, then just copy this function
    and modify.

    """

    app = sanic.Sanic(name="entrance-app", log_config=None)
    app.config.RESPONSE_TIMEOUT = 3600
    app.config.KEEP_ALIVE_TIMEOUT = 75

    # Websocket handling
    @app.websocket("/ws")
    async def handle_ws(request, ws):
        log.info("New websocket client")
        ws_handler = WebsocketHandler(ws, config["features"])
        await ws_handler.handle_incoming_requests()

    # Static file handling
    #
    # Note:
    # The order of 'app.static()' and route declarations (with '@app.route()')
    # matters here. For static to be a fallback it should come after route
    # declarations, e.g. if a route path contains a parameter.
    static_dir = file_location + config["start"]["static_dir"]
    app.static("/", static_dir)

    @app.route("/")
    async def home_page(request):
        return await sanic.response.file(static_dir + "/index.html")

    return app



# Netconf feature
#
# Copyright (c) 2018 Ensoft Ltd

from .tgt_base import TargetFeature


class NetconfFeature(TargetFeature):
    """
    Feature that exposes Netconf operations
    """

    #
    # Schema
    #
    name = "netconf"
    requests = {"netconf": ["op", "__req__"]}

    # We decode out the actual netconf op as follows
    op_requests = {
        "get": ["value"],
        "get_config": ["value"],
        "edit_config": ["value"],
        "commit": [],
        "validate": [],
        "discard_changes": [],
    }

    #
    # Implementation
    #
    async def connect(self, conn_factory):
        """
        Connect and get ready for the next commit request
        """
        self.connection = await conn_factory.get_netconf_connection("netconf")
        self.add_connection(self.connection, from_scratch=True)

    async def do_netconf(self, op, req):
        """
        Do one of the netconf operations specified in op_requests
        """
        try:
            fn = getattr(self.connection, op)
            args = [req[arg] for arg in self.op_requests[op]]
            rpc_reply = await fn(*args)
            result = str(rpc_reply)

            # Semi-hack for commit check:
            #
            # If a validate operation returns errors, but they all have
            # severity 'warning' not 'error', then consider the operation
            # a success
            if (
                op == "validate"
                and not rpc_reply.ok
                and not "<error-severity>error</error-severity>" in result
            ):
                return self._rpc_success(result)

            # Usual processing
            elif rpc_reply.ok:
                return self._rpc_success(result)
            else:
                return self._rpc_failure(result)

        except Exception as e:
            return self._rpc_failure(str(e))



# CLI config feature
#
# Copyright (c) 2018 Ensoft Ltd

from .tgt_base import TargetFeature


class CLIConfigFeature(TargetFeature):
    """
    Feature that enters configuration via the CLI
    """

    #
    # Schema
    #
    name = "cli_config"
    requests = {
        "cli_config_load": ["config"],
        "cli_config_commit": ["check_only"],
        "cli_config_get_failures": [],
        "cli_config_get_unsupported": [],
    }

    #
    # Implementation
    #
    async def connect(self, conn_factory):
        """
        Connect and get ready for the next commit request
        """
        self.connection = await conn_factory.get_cli_connection(
            "config", finalizer=self.finalizer
        )
        self.add_connection(self.connection, from_scratch=True)

    async def finalizer(self):
        """
        Finalize a new connection
        """
        await self.connection.send("configure\n", override=True)
        await self.connection.expect_prompt(override=True)

    async def do_cli_config_load(self, config):
        """
        Load up a config buffer with some CLI
        """
        # Clear any previous confiuration in the session
        await self.connection.send("clear\n")
        await self.connection.expect_prompt()

        # Enter the new configuration
        errors = []
        for line in config.split("\n"):
            await self.connection.send(line + "\n")
            result = await self.connection.expect_prompt()
            if len(result) > len(line) + 4:
                errors.append(result)

        # Check for parser-rejected syntax errors
        if len(errors) > 0:
            return self._rpc_failure("\n".join(errors))
        else:
            return self._rpc_success()

    async def do_cli_config_commit(self, check_only=False):
        """
        Commit a config buffer populated with cli_config_load
        """
        command = "validate commit" if check_only else "commit"
        await self.connection.send(command + "\n")
        result = await self.connection.expect_prompt()
        if len(result) > 200:
            return self._rpc_failure(result)
        else:
            return self._rpc_success()

    async def do_cli_config_get_failures(self):
        """
        Get config errors
        """
        await self.connection.send("show configuration failed\n")
        result = await self.connection.expect_prompt(strip_top=True)
        result = result.strip()
        if len(result):
            return self._rpc_failure(result)
        else:
            return self._rpc_success()

    async def do_cli_config_get_unsupported(self):
        """
        Get config items that are unsupported by validation
        """
        await self.connection.send("show configuration validation unsupported\n")
        result = await self.connection.expect_prompt(strip_top=True)
        result = result.strip()
        if result == "% No such configuration item(s)":
            return self._rpc_success()
        else:
            return self._rpc_failure(result)



# ConfiguredFeature base class
#
# Copyright (c) 2018 Ensoft Ltd

import logging

from .base import Feature
from ..exceptions import EntranceError

log = logging.getLogger(__name__)


class ConfiguredFeature(Feature):
    """
    A feature that is started by configuration, once per websocket
    """

    # Dictionary of configuration options and default values
    config = {}

    def __init__(self, ws_handler, config):
        """
        Merge in any specifed configuration with our defaults
        """
        super().__init__(ws_handler)
        for key, val in config.items():
            if key in self.config:
                self.config[key] = val
            else:
                msg_parts = [
                    'Invalid config item "{}" for feature {}'.format(key, self.name),
                    'Possible items are: {}'.format(sorted(self.config.keys())),
                ]
                log.critical(
                    "\n!!\n!!\n!! %s\n!!\n!!\n!!",
                    "\n!! ".join(msg_parts)
                )
                raise EntranceError(". ".join(msg_parts))

    @classmethod
    def all(cls):
        """
        List all subclasses
        """
        # for now assume a flat subclass hierarchy
        return cls.__subclasses__()



# Setup definition for EnTrance package
#
# To upload a new version to PyPi using nix:
#
# - `nix-shell -p 'python3.withPackages(p:[p.pip p.setuptools p.twine])'`
# - `python3 setup.py sdist`
# - If that all looks ok, `twine upload dist/*`
#
# I suspect there's something else that should be happening with wheel, but
# this seems to work.

import os, sys
from setuptools import find_packages, setup

# Router features require heavy dependencies, so include only when actually required
# via depending on 'entrance[with-router-features' rather than just 'entrance'.
router_feature_deps = ["janus", "ncclient", "paramiko"]

# Acutally, an icky second way of including the optional dependencies would just be
# to rewrite this from '[]' to 'router_feature_deps'. I'm looking at you, nix...
extra_deps = []

# Sanic dropped support for Python 3.5, leaving a long-term support version at 18.12.
# Otherwise, we need the new app loader introduced in 22.9.0.
v = sys.version_info
assert v.major == 3
assert v.minor >= 5
sanic = "sanic==18.12.0" if v.minor == 5 else "sanic>=22.9.0"

# ujson is an indirect dependency (via sanic) and version 2.x has a C part that doesn't
# compile.
ujson = "ujson!=2.*"

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="entrance",
    version="1.1.18",
    author="Ensoft Ltd",
    description="Server framework for web apps",
    url="https://github.com/ensoft/entrance",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=["pyyaml", sanic, ujson] + extra_deps,
    extras_require={"with-router-features": router_feature_deps},
)
